
/*
 * original:lejos-osek/src/nxtvm/platform/nxt/nxt_spi.c rev.1.1.1.1
 * modified history.
 * ??/??/200?: Modifier Andy Shaw
 *           : see Andy Shaw notes.
 *              added spi_set_mode and nxt_spi_refresh, used spi_isr_C.
 * 04/17/2010: Modifider Ryosuke Takeuchi
 *           : Modified Device Interface to ITRON-SIL interface and Changed
 *             AT91SAM7.h to SIL defined include file.
 *           : Deleted UNRE chack in twi_isr_C function.
 *           : Modified Interrupt disable in nxt_spi_init.
 */

#include "rtos.h"
#include "nxt_spi.h"
#include "byte_fifo.h"

/*
 * Note that this is not a normal SPI interface, 
 * it is a bodged version as used by the NXT's 
 * display.
 *
 * The display does not use MISO because you can
 * only write to it in serial mode.
 *
 * Instead, the MISO pin is not used by the SPI
 * and is instead driven as a PIO pin for controlling CD.
 *
 * Addional notes from Andy Shaw
 * The following code now contains the capability to perform display
 * updates using dma, This code was inspired by the nxos lcd/spi code
 * (Thanks guys). More details of nxos can be found at:
 * http://nxt.natulte.net/nxos/trac
 *
 */


#define CS_PIN	(1<<10)
#define CD_PIN  (1<<12)
const U8 *display = (U8 *) 0;
volatile U8 dirty = 0;
volatile U8 page = 0;
volatile const U8 *data = (U8 *) 0;
U8 mode = 0xff;

extern void spi_isr_entry(void);

static void spi_set_mode(U8 m)
{
  U32 status;

  /* nothing to do if we are already in the correct mode */
  if (m == mode) return;

  /* Wait until all bytes have been sent */
  do {
    status = sil_rew_mem((VP)(TADR_SPI_BASE+TOFF_SPI_SR));
  } while (!(status & 0x200));
  /* Set command or data mode */
  if (m)
    sil_wrw_mem((VP)(TADR_PIO_BASE+TOFF_PIO_SODR), CD_PIN);
  else
    sil_wrw_mem((VP)(TADR_PIO_BASE+TOFF_PIO_CODR), CD_PIN);
  /* remember the current mode */
  mode = m;
}


void
spi_isr_C(void)
{
  if (page == 0)
  {
    /* Check to see if we have data to display */
    if (dirty != 0)
    {
      data = display;
      dirty = 0;
    }
    else
    {
      /* No so turn things off. It will get re-set if we ever have anything
         to display
      */
      sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_IDR), SPI_ENDTX);
      return;
    }
  }
  /* Make sure we are in data mode */
  spi_set_mode(1);
  /* now do the transfer. We make use of the auto-wrap function so simply
   * need to send 8*132 bytes to get back to where we started. However the
   * display buffer is structured as series of 100 byte lines, so we need to
   * get tricky. I've made the display one line longer (9 lines) and so when we
   * send the data we send 100 bytes from the actual line plus 32 padding bytes
   * (that are not actually seen), from the next line. The extra line means
   * that this is safe to do. If we can redefine the display as a 8*132 then
   * we could just use a single dma transfer (instead of 8, 132 byte ones).
   * However I'm not sure if this would be safe.
   */
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_TNPR), (U32)data);
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_TNCR), 132);
  page = (page + 1) % 8;
  data += 100;
}


void
nxt_spi_init(void)
{
#define OSC 48054805
#define SPI_BITRATE 2000000
  unsigned long pbase = TADR_PIO_BASE;
  sil_wrw_mem((VP)(TADR_PMC_BASE+TOFF_PMC_PCER), 1<<IRQ_SPI_PID);	/* Enable MCK clock     */
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PER), AT91S_PIO_PA12);	/*EnableA0onPA12*/
  sil_wrw_mem((VP)(pbase+TOFF_PIO_OER), AT91S_PIO_PA12);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_CODR), AT91S_PIO_PA12);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PDR), AT91S_PIO_PA14);	/*EnableSPCKonPA14*/
  sil_wrw_mem((VP)(pbase+TOFF_PIO_ASR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_ODR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_OWER), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_MDDR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PUDR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_IFDR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_CODR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_IDR), AT91S_PIO_PA14);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PDR), AT91S_PIO_PA13);	/*EnablemosionPA13*/
  sil_wrw_mem((VP)(pbase+TOFF_PIO_ASR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_ODR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_OWER), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_MDDR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PUDR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_IFDR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_CODR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_IDR), AT91S_PIO_PA13);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PDR), AT91S_PIO_PA10);	/*Enablenpcs0onPA10*/
  sil_wrw_mem((VP)(pbase+TOFF_PIO_BSR), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_ODR), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_OWER), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_MDDR), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_PUDR), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_IFDR), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_CODR), AT91S_PIO_PA10);
  sil_wrw_mem((VP)(pbase+TOFF_PIO_IDR), AT91S_PIO_PA10);

  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_CR), SPI_SWRST);	/*Softreset*/
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_CR), SPI_SPIEN);	/*Enablespi*/
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_MR), SPI_MSTR|SPI_MODFDIS | (0xB<<16));
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_CSR2), ((OSC/SPI_BITRATE)<<8) | SPI_CPOL);

  /* Set mode to unknown */
  mode = 0xff;

  /* Set up safe dma refresh state */
  data = display = (U8 *) 0;
  dirty = 0;
  page = 0;
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_PTCR), PDC_TXTEN);
}

void
nxt_spi_write(U32 CD, const U8 *data, U32 nBytes)
{
  U32 status;
  U32 cd_mask = (CD ? 0x100 : 0);

  spi_set_mode(CD);
  while (nBytes) {
    sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_TDR), *data | cd_mask);
    data++;
    nBytes--;
    /* Wait until byte sent */
    do{
      status = sil_rew_mem((VP)(TADR_SPI_BASE+TOFF_SPI_SR));
    }while(!(status & 0x200));
  }
}

void
nxt_spi_set_display(const U8 *disp)
{
  /* Set the display buffer to be used for dma refresh.
   * it is really only safe to set the display once. Should probably
   * sort this out so that it is set separately from requesting a refresh
   */
  if (!display) display = disp;
}

void
nxt_spi_refresh(void)
{
  /* Request the start of a dma refresh of the display 
   */
  // If the display is not set nothing to do.
  if (!display) return;
  // Say we have changes
  dirty = 1;
  sil_wrw_mem((VP)(TADR_SPI_BASE+TOFF_SPI_IER), SPI_ENDTX);
}
